from lightwood.encoders.encoder_base import BaseEncoder
from lightwood.encoders.datetime.datetime import DatetimeEncoder
from lightwood.encoders.image.img_2_vec import Img2VecEncoder
from lightwood.encoders.image.nn import NnAutoEncoder
from lightwood.encoders.numeric.numeric import NumericEncoder
from lightwood.encoders.numeric.ts_numeric import TsNumericEncoder
from lightwood.encoders.text.short import ShortTextEncoder
from lightwood.encoders.text.vocab import VocabularyEncoder
from lightwood.encoders.text.rnn import RnnEncoder as TextRnnEncoder
from lightwood.encoders.categorical.onehot import OneHotEncoder
from lightwood.encoders.categorical.autoencoder import CategoricalAutoEncoder
from lightwood.encoders.time_series.rnn import TimeSeriesEncoder as TsRnnEncoder
from lightwood.encoders.time_series.plain import TimeSeriesPlainEncoder
from lightwood.encoders.categorical.multihot import MultiHotEncoder
from lightwood.encoders.text.pretrained import PretrainedLang


class DateTime:
    DatetimeEncoder = DatetimeEncoder


class Image:
    Img2VecEncoder = Img2VecEncoder
    NnAutoEncoder = NnAutoEncoder


class Numeric:
    NumericEncoder = NumericEncoder
    TsNumericEncoder = TsNumericEncoder


class Text:
    ShortTextEncoder = ShortTextEncoder
    TextRnnEncoder = TextRnnEncoder
    VocabularyEncoder = VocabularyEncoder
    PretrainedLang = PretrainedLang

class Categorical:
    OneHotEncoder = OneHotEncoder
    CategoricalAutoEncoder = CategoricalAutoEncoder

class TimeSeries:
    TsRnnEncoder = TsRnnEncoder
    TimeSeriesPlainEncoder = TimeSeriesPlainEncoder

class BuiltinEncoders:
    DateTime = DateTime
    Image = Image
    Numeric = Numeric
    Text = Text
    Categorical = Categorical
    TimeSeries = TimeSeries
    # Audio = Audio


BUILTIN_ENCODERS = BuiltinEncoders



import torch
from torch import nn
from lightwood.mixers.helpers.default_net import DefaultNet
from lightwood.helpers.torch import LightwoodAutocast
from itertools import chain


class ArNet(DefaultNet):
    """
    DefaultNet variant that adds a secondary stream (simple linear layer) with constrained
    weights to learn autoregressive coefficients for numerical time series targets
    """
    def __init__(self, dynamic_parameters,
                 transformer,
                 input_size=None,
                 output_size=None,
                 nr_outputs=None,
                 shape=None,
                 max_params=3e5,
                 dropout=None,
                 pretrained_net=None):
        """
        :param data_info: lightwood.mixers.helpers.transformer.Transformer;
        used to save the indexes for which we learn the autoregressive branch of the network
        """
        assert len(transformer.output_features) == 1  # ArNet supports single target prediction only
        super().__init__(dynamic_parameters,
                         input_size=input_size,
                         output_size=output_size,
                         nr_outputs=nr_outputs,
                         shape=shape,
                         max_params=max_params,
                         dropout=dropout,
                         pretrained_net=pretrained_net
                         )
        target = transformer.output_features[0]
        self.transformer = transformer
        self.ar_column = f'__mdb_ts_previous_{target}'
        self.ar_idxs = list(*[transformer.input_idxs[col]
                              for col in transformer.input_idxs
                              if col == self.ar_column])

        dims = [(len(self.ar_idxs), transformer.feature_len_map[target])]
        linears = [nn.Linear(in_features=inf, out_features=outf) for inf, outf in dims]
        self.ar_net = nn.Sequential(*linears)
        self.ar_net.to(self.device)

    def to(self, device=None, available_devices=None):
        self.ar_net.to(device)
        return super().to(device, available_devices)

    def forward(self, input):
        with LightwoodAutocast():
            residual_output = self._foward_net(input)
            ar_output = self.ar_net(input[:, self.ar_idxs])
            # force unit root
            if self.ar_net.training:
                self.ar_net._modules['0'].weight = nn.Parameter(torch.clamp(self.ar_net._modules['0'].weight,
                                                                            0.0,
                                                                            0.999))

        return ar_output + residual_output



from lightwood.encoders.text.pretrained import PretrainedLang
from lightwood.encoders.text.rnn import RnnEncoder
from lightwood.encoders.text.tfidf import TfidfEncoder
from lightwood.encoders.text.short import ShortTextEncoder
from lightwood.encoders.text.vocab import VocabularyEncoder
